/*****************************************************************************/

/*
 *	ds14xxe.c  -- Dallas Semiconductor One Wire driver for
 *	DS1410E/DS1405E parallel adapters.  Also may work with
 *	DS1410D/DS1405D.
 *
 *	NOTE: Do not contact Dallas Semiconductor for support with
 *	this driver. 
 *
 *	Copyright (C) 1997  Bryan Armstrong (bryana@mail.airmail.net).
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*****************************************************************************/

/*
 * Version History
 * 1.01.0 - Fixed off by one bug when converting legacy port number
 *          to table offset.
 *          4/17/98
 * 1.00.0 - First public release.  Made driver backwards compatible
 *          with legacy one-wire routines (Setup()).
 * 0.90.0 - Test
 */

/* The config.h include must be first! */
#include <linux/config.h>

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <linux/ioport.h>  /* For inp/outp, etc. */
#include <linux/delay.h>   /* For udelay */

#include "ds14xxe.h"

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

/* Here's the info for the ports */
struct dow_struct dow_table[] =
{ { 0x3bc },
  { 0x378 },
  { 0x278 } };

#define	DOW_NO			3
#define MAX_DS1481_BLKSIZE	32
#define BUSYTIMEOUT		4000
#define MINTIMEOUT		50

static int dow_major;
static unsigned char LegacyPortNumber = 0;

struct dow_buffer
{
  unsigned char DS1481Command;
  unsigned char DS1481Buffer[MAX_DS1481_BLKSIZE]; 
} DOWComm[DOW_NO];

/*
Commands such as Reset, One Bit, and Zero Bit
go through this function.
*/
static int do_ds1481command(char command,int minor)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  
  int loopcount;
  
  /* Save the base register */
  base = inb_p(DOW_BASE(minor));

  /* Output the command */
  outb_p(command,DOW_BASE(minor));
  /* Get the control byte */
  control = inb_p(DOW_CONTROL(minor));
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));

  /* Take ENI low */
  control |= DS1481_ENI;
  outb_p(control,DOW_CONTROL(minor));

  /* Wait for BSY to go low */
  loopcount = 0;
  do
  {
    udelay(1);
    status = inb_p(DOW_STATUS(minor));
    loopcount++;
  }
  while (!(status & DS1481_O1BSY) && (status & DS1481_O2BSY) && (loopcount < MINTIMEOUT));

  /* Bring DCLK high. */
  outb_p(0xFF,DOW_BASE(minor));

  /* Wait for BSY to go high */
  loopcount = 0;
  do
  {
    udelay(1);
    status = inb_p(DOW_STATUS(minor));
    loopcount++;
    /* Every 100us or so, give up the CPU so something else can run. */
    if ((loopcount % 100) == 0)
      schedule();
  }
  while ((status & DS1481_O1BSY) && !(status & DS1481_O2BSY) && (loopcount < BUSYTIMEOUT));  

  /* Clock in the result of the command */
  outb_p(0xFE,DOW_BASE(minor));
  udelay(2);
  result = inb_p(DOW_STATUS(minor));
  outb_p(0xFF,DOW_BASE(minor));

  result = ((result ^ 0x80) & 0x90) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));
  outb_p(base,DOW_BASE(minor));

  return(result);
}

/*
Toggle Overdrive function.
*/
static int do_ds1481toggleOD(int minor)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  
  /* Save the base register */
  base = inb_p(DOW_BASE(minor));

  /* Output the command */
  outb_p(0xFC,DOW_BASE(minor));
  /* Get the control byte */
  control = inb_p(DOW_CONTROL(minor));
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));

  /* Take ENI low */
  control |= DS1481_ENI;
  outb_p(control,DOW_CONTROL(minor));

  /* Wait for Overdrive status to become available */
  udelay(4);

  /* Get OD status */
  result = inb_p(DOW_STATUS(minor));
  result = ((result ^ 0x80) & 0x90) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));
  outb_p(base,DOW_BASE(minor));

  return(result);
}

/*
Brick Check Function
*/
static int do_ds1481checkbrick(int minor)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  int loopcount;
  
  /* Save the base register */
  base = inb_p(DOW_BASE(minor));

  /* Output the Reset command */
  outb_p(0xFD,DOW_BASE(minor));
  /* Get the control byte */
  control = inb_p(DOW_CONTROL(minor));
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));

  /* Take ENI low */
  control |= DS1481_ENI;
  outb_p(control,DOW_CONTROL(minor));

  /* Wait for BSY to go low */
  loopcount = 0;
  do
  {
    udelay(1);
    status = inb_p(DOW_STATUS(minor));
    loopcount++;
  }
  while (!(status & DS1481_O1BSY) && (status & DS1481_O2BSY) && (loopcount < MINTIMEOUT));
  
  result = (loopcount < MINTIMEOUT) ? 1 : 0;

  /* Wait for BSY to go high */
  loopcount = 0;
  do
  {
    udelay(1);
    status = inb_p(DOW_STATUS(minor));
    loopcount++;
    /* Every 100us or so, give up the CPU so something else can run. */
    if ((loopcount % 100) == 0)
      schedule();
  }
  while ((status & DS1481_O1BSY) && !(status & DS1481_O2BSY) && (loopcount < BUSYTIMEOUT));  

  /* If we have to wait too long the part is not there, return 0 */
  if (result)
    result = (loopcount < BUSYTIMEOUT) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  outb_p(control,DOW_CONTROL(minor));
  outb_p(base,DOW_BASE(minor));

  return(result);
}

static int lseek_dow(struct inode *inode,struct file *file,off_t offset, int orig)
{
  return(file->f_pos = 0);
}

/*
Reads the results of the one wire command out of the buffer
*/
static int read_dow(struct inode *inode,struct file *file,char *buf,int count)
{
  int returnval;
  int minsize;
  unsigned int minor = MINOR(inode->i_rdev);
  
  returnval = verify_area(VERIFY_WRITE,buf,count);
  if (returnval != 0) 
    return(returnval);
  minsize = (count < MAX_DS1481_BLKSIZE)?count:MAX_DS1481_BLKSIZE;
  memcpy_tofs(buf,(char *)DOWComm[minor].DS1481Buffer,minsize);

  return(minsize);
}

/*
Completes the actual one wire command
*/
static int write_dow(struct inode *inode,struct file *file,char *buf,int count)
{
  int returnval;
  int minsize;
  unsigned int minor = MINOR(inode->i_rdev);
  int bitcount;
  int bytecount;
  unsigned char mask;
  unsigned char portindex;
  
  /* If we're not in legacy parallel mode, use
     the minor number to get the port address, otherwise
     use the LegacyPortNumber that was set earlier by
     a reset call.
  */
  if (LegacyPortNumber == 0)
    portindex = minor;
  else
    portindex = LegacyPortNumber - 1;

  /* Get command byte */
  returnval = verify_area(VERIFY_READ,buf,count);
  if (returnval != 0) 
    return(returnval);
  minsize = (count < MAX_DS1481_BLKSIZE+1)?count:MAX_DS1481_BLKSIZE+1;
  memcpy_fromfs((char *)&DOWComm[minor].DS1481Command,buf,minsize);
  switch (DOWComm[minor].DS1481Command)
  {
    case DO_RESET:
      /* If we get two bytes in the reset command, the user is
         wanting to do legacy parallel communication.  Use the
         second byte as the port number to work with instead of
         using the minor number.  If the user then later sends only
         one byte, we go back into normal mode.
      */
      if (minsize == 2)
      {
        /* If the port number passed is between 1 and 3
           it's good (LPT1-3)
        */
        if ((DOWComm[minor].DS1481Buffer[0] > 0) &&
            (DOWComm[minor].DS1481Buffer[0] < 4))
        {
          LegacyPortNumber = DOWComm[minor].DS1481Buffer[0];
          portindex = LegacyPortNumber - 1; /* Index starts at zero. */
        }
        else
        {
          LegacyPortNumber = 0;
          portindex = minor;
        }
      }
      else
      {
        LegacyPortNumber = 0;
        portindex = minor;
      }
        
      returnval = do_ds1481command(0xFD,portindex);

      DOWComm[minor].DS1481Buffer[0] = returnval ^ 0x01;
      break;
    case DO_BIT:
      /* Get bit to send. */
      if (DOWComm[minor].DS1481Buffer[0] & 0x01)
        returnval = do_ds1481command(0xFF,portindex);
      else
        returnval = do_ds1481command(0xFE,portindex);
      DOWComm[minor].DS1481Buffer[0] = returnval;
      break;
    case DO_BYTE:
      mask = 0x01;
      for (bitcount = 0;bitcount < 8;bitcount++)
      {
        if (DOWComm[minor].DS1481Buffer[0] & mask)
          returnval = do_ds1481command(0xFF,portindex);
        else
          returnval = do_ds1481command(0xFE,portindex);
        if (returnval & 0x01)
          DOWComm[minor].DS1481Buffer[0] |= mask;
        else
          DOWComm[minor].DS1481Buffer[0] &= ~mask;
        mask <<= 1;
      }
      break;
    case DO_BLOCK:
      /* Get block of bytes to send. */
//      printk("min: %d\n",minsize);
      for (bytecount = 0;bytecount < (minsize-1);bytecount++)
      {
//        printk("Byte: %X\n",DOWComm[minor].DS1481Buffer[bytecount]);
        mask = 0x01;
        for (bitcount = 0;bitcount < 8;bitcount++)
        {
          if (DOWComm[minor].DS1481Buffer[bytecount] & mask)
            returnval = do_ds1481command(0xFF,portindex);
          else
            returnval = do_ds1481command(0xFE,portindex);
          if (returnval & 0x01)
            DOWComm[minor].DS1481Buffer[bytecount] |= mask;
          else
            DOWComm[minor].DS1481Buffer[bytecount] &= ~mask;
          mask <<= 1;
        }
        /* Schedule between bytes sent. Give up some time. */
        schedule();
      }
      break;
    case TOGGLE_OVERDRIVE:
      DOWComm[minor].DS1481Buffer[0] = do_ds1481toggleOD(portindex);
      break;
    case TOGGLE_PASSTHRU:
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      break;
    case CHECK_BRICK:
      DOWComm[minor].DS1481Buffer[0] = do_ds1481checkbrick(portindex);
      break;
    default:
      printk("Bad Command\n");
      break;
  }

  return(minsize);
}

static int open_dow(struct inode *inode,struct file *file)
{
  MOD_INC_USE_COUNT;
  return(0);
}

static void release_dow(struct inode *inode,struct file *file)
{
  MOD_DEC_USE_COUNT;
}

static struct file_operations dow_fops = 
{
  lseek_dow,
  read_dow,
  write_dow,
  NULL,
  NULL,
  NULL,
  NULL,
  open_dow,
  release_dow,
  NULL,
  NULL,
  NULL,
  NULL
};

#ifndef MODULE
long dow_init(long mem_start, long mem_end)
{
  if (dow_major = register_chrdev(0,"dow",&dow_fops))
    printk("unable to get major for dow device\n");
#ifdef DOW_DEBUG
  else
    printk("DOW driver got major number: %d\n",dow_major);
#endif  
  return(mem_start);
}
#else
int init_module(void)
{
  if ((dow_major = register_chrdev(0,"dow",&dow_fops)) == -EBUSY)
  {
    printk("unable to get major for dow device\n");
    return(-EIO);
  }
#ifdef DOW_DEBUG
  else
    printk("DOW driver got major number: %d\n",dow_major);
#endif  
  return(0);
}

void cleanup_module(void)
{
  unregister_chrdev(dow_major,"dow");
}
#endif

