/*****************************************************************************/

/*
 *	ds14xxe.c  -- Dallas Semiconductor One Wire driver for
 *	DS1410E/DS1405E parallel adapters.  Also may work with
 *	DS1410D/DS1405D.
 *
 *	NOTE: Do not contact Dallas Semiconductor for support with
 *	this driver. 
 *
 *	Copyright (C) 1997  Bryan Armstrong (bryana@mail.airmail.net).
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*****************************************************************************/

/*
 * Version History
 * 1.10.0 - Removed non-module code.  Will only work as a module now.
 *          Updated calls to use Linux kernel 2.2 convention.
 *          Updated inode referencing to 2.2 convention.
 *          Updated user<->kernel data passing to 2.2 convention.
 *          2/1/99
 * 1.01.0 - Fixed off by one bug when converting legacy port number
 *          to table offset.
 *          4/17/98
 * 1.00.0 - First public release.  Made driver backwards compatible
 *          with legacy one-wire routines (Setup()).
 * 0.90.0 - Test
 */

/* The config.h include must be first! */
//#include <linux/config.h>

#include <linux/types.h>
#include <linux/fs.h> /* For struct file_operations */
//#include <linux/mm.h>
#include <linux/errno.h>
//#include <asm/segment.h>
#include <asm/io.h>
#include <asm/uaccess.h> /* For put_user, get_user */
#include <linux/kernel.h> 
#include <linux/ioport.h>  /* For inp/outp, etc. */
#include <linux/delay.h>   /* For udelay */

#include "ds14xxe.h"

#include <linux/module.h> /* For module stuff */
#include <linux/version.h>

#include <linux/proc_fs.h> /* For proc support */

#include <linux/parport.h> /* For parport support */

struct dow_struct
{
  int base;
};

/* Here's the info for the ports */
struct dow_struct dow_table[] =
{ { 0x3bc },
  { 0x378 },
  { 0x278 } };

/* Here's the info for the parport sharing. */
struct dow_parport_struct
{
  struct pardevice *dev;
};

struct dow_parport_struct dow_parport_table[] =
{
  { NULL },
  { NULL },
  { NULL }
};

#define	DOW_NO			3
#define MAX_DS1481_BLKSIZE	32
#define BUSYTIMEOUT		4000
#define MINTIMEOUT		50

static int dow_major;
static unsigned char LegacyPortNumber = 0;

static int numresets = 0;
static int numbytes = 0;
static int numbits = 0;
static int numblocks = 0;
static int numoverdrives = 0;
static int numpassthroughs = 0;
static int numcheckbricks = 0;
static int numbadcommands = 0;

struct dow_buffer
{
  unsigned char DS1481Command;
  unsigned char DS1481Buffer[MAX_DS1481_BLKSIZE]; 
} DOWComm[DOW_NO];

/*
Commands such as Reset, One Bit, and Zero Bit
go through this function.
*/
static int do_ds1481command(char command,int portindex)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  
  int loopcount;
  
  /* Save the base register */
  base = 0xFF;
//  base = READ_DATA(portindex);

  /* Output the command */
  WRITE_DATA(command,portindex);
  /* Get the control byte */
  control = READ_CONTROL(portindex);
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);

  /* Take ENI low */
  control |= DS1481_ENI;
  WRITE_CONTROL(control,portindex);

  /* Wait for BSY to go low */
  loopcount = 0;
  do
  {
    udelay(1);
    status = READ_STATUS(portindex);
    loopcount++;
  }
  while (!(status & DS1481_O1BSY) && (status & DS1481_O2BSY) && (loopcount < MINTIMEOUT));

  /* Bring DCLK high. */
  WRITE_DATA(0xFF,portindex);

  /* Wait for BSY to go high */
  loopcount = 0;
  do
  {
    udelay(1);
    status = READ_STATUS(portindex);
    loopcount++;
    /* Every 100us or so, give up the CPU so something else can run. */
    if ((loopcount % 100) == 0)
      schedule();
  }
  while ((status & DS1481_O1BSY) && !(status & DS1481_O2BSY) && (loopcount < BUSYTIMEOUT));  

  /* Clock in the result of the command */
  WRITE_DATA(0xFE,portindex);
  udelay(2);
  result = READ_STATUS(portindex);
  WRITE_DATA(0xFF,portindex);

  result = ((result ^ 0x80) & 0x90) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);
  WRITE_DATA(base,portindex);

  return(result);
}

/*
Toggle Overdrive function.
*/
static int do_ds1481toggleOD(int portindex)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  
  /* Save the base register */
  base = 0xFF;
//  base = READ_DATA(portindex);

  /* Output the command */
  WRITE_DATA(0xFC,portindex);
  /* Get the control byte */
  control = READ_CONTROL(portindex);
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);

  /* Take ENI low */
  control |= DS1481_ENI;
  WRITE_CONTROL(control,portindex);

  /* Wait for Overdrive status to become available */
  udelay(4);

  /* Get OD status */
  result = READ_STATUS(portindex);
  result = ((result ^ 0x80) & 0x90) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);
  WRITE_DATA(base,portindex);

  return(result);
}

/*
Brick Check Function
*/
static int do_ds1481checkbrick(int portindex)
{
  unsigned char base;
  unsigned char control;
  unsigned char status;
  unsigned char result;
  int loopcount;
  
  /* Save the base register */
  base = 0xFF;
//  base = READ_DATA(portindex);

  /* Output the Reset command */
  WRITE_DATA(0xFD,portindex);
  /* Get the control byte */
  control = READ_CONTROL(portindex);
  /* Take ENI high */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);

  /* Take ENI low */
  control |= DS1481_ENI;
  WRITE_CONTROL(control,portindex);

  /* Wait for BSY to go low */
  loopcount = 0;
  do
  {
    udelay(1);
    status = READ_STATUS(portindex);
    loopcount++;
  }
  while (!(status & DS1481_O1BSY) && (status & DS1481_O2BSY) && (loopcount < MINTIMEOUT));
  
  result = (loopcount < MINTIMEOUT) ? 1 : 0;

  /* Wait for BSY to go high */
  loopcount = 0;
  do
  {
    udelay(1);
    status = READ_STATUS(portindex);
    loopcount++;
    /* Every 100us or so, give up the CPU so something else can run. */
    if ((loopcount % 100) == 0)
      schedule();
  }
  while ((status & DS1481_O1BSY) && !(status & DS1481_O2BSY) && (loopcount < BUSYTIMEOUT));  

  /* If we have to wait too long the part is not there, return 0 */
  if (result)
    result = (loopcount < BUSYTIMEOUT) ? 1 : 0;

  /* Reset the state of everything */
  control &= ~(DS1481_ENI);
  WRITE_CONTROL(control,portindex);
  WRITE_DATA(base,portindex);

  return(result);
}

static loff_t lseek_dow(struct file *file,loff_t offset, int orig)
{
  return(file->f_pos = 0);
}

/*
Reads the results of the one wire command out of the buffer
*/
static ssize_t read_dow(struct file *file,char *buf,size_t count,loff_t *ppos)
{
  int i;
  int returnval;
  size_t minsize;
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
    
  if (count < 0)
    return (-EIO);
    
  minsize = (count < MAX_DS1481_BLKSIZE)?count:MAX_DS1481_BLKSIZE;

  for (i = 0;i < minsize;i++)
    put_user(DOWComm[minor].DS1481Buffer[i],buf+i);
      
  return(minsize);
}

/*
Completes the actual one wire command
*/
static ssize_t write_dow(struct file *file,char *buf,size_t count,loff_t *ppos)
{
  int i;
  int returnval;
  size_t minsize;
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
  
  int bitcount;
  int bytecount;
  unsigned char mask;
  unsigned char portindex;
  
  /* If we're not in legacy parallel mode, use
     the minor number to get the port address, otherwise
     use the LegacyPortNumber that was set earlier by
     a reset call.
  */
  if (LegacyPortNumber == 0)
    portindex = minor;
  else
    portindex = LegacyPortNumber - 1;

  /* Check for negative length */
  if (count < 0)
    return (-EIO);

#ifdef DOW_USE_PARPORT
  if (dow_parport_table[portindex].dev == NULL)
  {
#ifdef DOW_DEBUG
printk("Entry was NULL %d\n",portindex);
//    printk("Entry %d was NULL\n",portindex);
#endif
    return (-EIO);
  }

  if (parport_claim(dow_parport_table[portindex].dev) != 0)
  {
    printk("Couldn't claim port %d\n",portindex);
    return (-EAGAIN);
  }
//  printk("Claimed port %d\n",portindex);
#endif

  minsize = (count < MAX_DS1481_BLKSIZE+1)?count:MAX_DS1481_BLKSIZE+1;

  /* Get command byte */
  if (minsize > 0)
    get_user(DOWComm[minor].DS1481Command,buf);

  /* Get command data */
  for (i = 1;i < minsize;i++)
    get_user(DOWComm[minor].DS1481Buffer[i-1],buf+i);

  switch (DOWComm[minor].DS1481Command)
  {
    case DO_RESET:
      numresets++;

      /* If we get two bytes in the reset command, the user is
         wanting to do legacy parallel communication.  Use the
         second byte as the port number to work with instead of
         using the minor number.  If the user then later sends only
         one byte, we go back into normal mode.
      */
      if (minsize == 2)
      {

#ifdef DOW_USE_PARPORT
        if (dow_parport_table[portindex].dev == NULL)
        {
#ifdef DOW_DEBUG
          printk("Error 3Entry was NULL\n");
#endif
        }
        else
        {
          parport_release(dow_parport_table[portindex].dev);
//          printk("Released port %d\n",portindex);
        }
#endif

        /* If the port number passed is between 1 and 3
           it's good (LPT1-3)
        */
        if ((DOWComm[minor].DS1481Buffer[0] > 0) &&
            (DOWComm[minor].DS1481Buffer[0] < 4))
        {
          LegacyPortNumber = DOWComm[minor].DS1481Buffer[0];
          portindex = LegacyPortNumber - 1; /* Index starts at zero. */
        }
        else
        {
          LegacyPortNumber = 0;
          portindex = minor;
        }

#ifdef DOW_USE_PARPORT
        if (dow_parport_table[portindex].dev == NULL)
        {
#ifdef DOW_DEBUG
          printk("Entry3 %d was NULL\n",portindex);
#endif
          DOWComm[minor].DS1481Buffer[0] = 0;
          return (-EIO);
        }

        if (parport_claim(dow_parport_table[portindex].dev) != 0)
        {
          printk("Couldn't claim port %d\n",portindex);
          return (-EAGAIN);
        }
//        printk("Claimed port %d\n",portindex);
#endif
      }
      else
      {
        LegacyPortNumber = 0;
        portindex = minor;
      }

      returnval = do_ds1481command(0xFD,portindex);

      DOWComm[minor].DS1481Buffer[0] = returnval ^ 0x01;
      break;
    case DO_BIT:
      numbits++;

      /* Get bit to send. */
      if (DOWComm[minor].DS1481Buffer[0] & 0x01)
        returnval = do_ds1481command(0xFF,portindex);
      else
        returnval = do_ds1481command(0xFE,portindex);
      DOWComm[minor].DS1481Buffer[0] = returnval;
      break;
    case DO_BYTE:
      numbytes++;
      
      mask = 0x01;
      for (bitcount = 0;bitcount < 8;bitcount++)
      {
        if (DOWComm[minor].DS1481Buffer[0] & mask)
          returnval = do_ds1481command(0xFF,portindex);
        else
          returnval = do_ds1481command(0xFE,portindex);
        if (returnval & 0x01)
          DOWComm[minor].DS1481Buffer[0] |= mask;
        else
          DOWComm[minor].DS1481Buffer[0] &= ~mask;
        mask <<= 1;
      }
      break;
    case DO_BLOCK:
      numblocks++;

      /* Get block of bytes to send. */
//      printk("min: %d\n",minsize);
      for (bytecount = 0;bytecount < (minsize-1);bytecount++)
      {
//        printk("Byte: %X\n",DOWComm[minor].DS1481Buffer[bytecount]);
        mask = 0x01;
        for (bitcount = 0;bitcount < 8;bitcount++)
        {
          if (DOWComm[minor].DS1481Buffer[bytecount] & mask)
            returnval = do_ds1481command(0xFF,portindex);
          else
            returnval = do_ds1481command(0xFE,portindex);
          if (returnval & 0x01)
            DOWComm[minor].DS1481Buffer[bytecount] |= mask;
          else
            DOWComm[minor].DS1481Buffer[bytecount] &= ~mask;
          mask <<= 1;
        }
        /* Schedule between bytes sent. Give up some time. */
        schedule();
      }
      break;
    case TOGGLE_OVERDRIVE:
      numoverdrives++;
      
      DOWComm[minor].DS1481Buffer[0] = do_ds1481toggleOD(portindex);
      break;
    case TOGGLE_PASSTHRU:
      numpassthroughs++;
      
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      do_ds1481toggleOD(portindex);
      break;
    case CHECK_BRICK:
      numcheckbricks++;
      
      DOWComm[minor].DS1481Buffer[0] = do_ds1481checkbrick(portindex);
      break;
    default:
      numbadcommands++;
      
      printk("Bad Command\n");
      break;
  }

#ifdef DOW_USE_PARPORT
  if (dow_parport_table[portindex].dev == NULL)
  {
#ifdef DOW_DEBUG
    printk("2Entry was NULL\n");
#endif
  }
  else
  {
    parport_release(dow_parport_table[portindex].dev);
//    printk("Released port %d\n",portindex);
  }
#endif

  return(minsize);
}

static int open_dow(struct inode *inode,struct file *file)
{
  MOD_INC_USE_COUNT;
  return(0);
}

static void release_dow(struct inode *inode,struct file *file)
{
  MOD_DEC_USE_COUNT;
}

static struct file_operations dow_fops = 
{
  lseek_dow,
  read_dow,
  write_dow,
  NULL, /* readdr */
  NULL, /* poll */
  NULL, /* ioctl */
  NULL, /* mmap */
  open_dow,
  NULL, /* flush */
  release_dow
};

/*************************************************************/
/* Begin proc filesystem code                                */
/*************************************************************/

struct proc_dir_entry *proc_dow;

static int procread_dow(char *buffer, char **start, off_t offset, 
                        int size, int *eof, void *data)
{
  int len = 0;
  
  /* FIXME: This could cause a buffer overrun.  YOU HAVE BEEN WARNED! */
//  printk("Length is %d, offset is %d\n",size,offset);
//  printk("Buffer is %d, start is %d\n",buffer,*start);
  len = sprintf(buffer,"DS14xxE Command Counters:\n");
  len += sprintf(buffer+len,"             Reset: %d\n",numresets);
  len += sprintf(buffer+len,"               Bit: %d\n",numbits);
  len += sprintf(buffer+len,"              Byte: %d\n",numbytes);
  len += sprintf(buffer+len,"             Block: %d\n",numblocks);
  len += sprintf(buffer+len,"         Overdrive: %d\n",numoverdrives);
  len += sprintf(buffer+len,"Toggle Passthrough: %d\n",numpassthroughs);
  len += sprintf(buffer+len,"      Bad Commands: %d\n",numbadcommands);
  len += sprintf(buffer+len,"\n");

// Do I need to set the start position???
//  *start = &buffer[0];

  return len;
}

/*************************************************************/
/* Begin module init/cleanup code                            */
/*************************************************************/

int init_module(void)
{
  struct parport *myenum;
  struct pardevice *mydevice;
  int i;
  
  int retval;
  
  if ((dow_major = register_chrdev(0,"dow",&dow_fops)) == -EBUSY)
  {
    printk("unable to get major number for dow device\n");
    return(-EIO);
  }
#ifdef DOW_DEBUG
  else
    printk("DOW driver got major number: %d\n",dow_major);
#endif  

  if (proc_dow = create_proc_entry("dow", S_IFREG | S_IRUGO | S_IWUSR, 0))
  {
    proc_dow -> read_proc = procread_dow;
  }
  else
  {
    printk("Unable to register /proc entry, unloading module.\n");
    unregister_chrdev(dow_major,"dow");
    return(-EIO);
  }

#ifdef DOW_USE_PARPORT
  myenum = parport_enumerate();
  i = 0;
  while (myenum != NULL)
  {
#ifdef DOW_DEBUG
    printk("Base address: %04X\n",myenum->base);
#endif
    mydevice = parport_register_device(myenum,"dow",NULL,NULL,NULL,0,NULL);
    dow_parport_table[i].dev = mydevice;
    
    i++;
    myenum = myenum->next;
  }
#endif
  

  return(0);
}

void cleanup_module(void)
{
  int i;
  
  remove_proc_entry("dow",0);
  unregister_chrdev(dow_major,"dow");
  
#ifdef DOW_USE_PARPORT
  for (i = 0;i < DOW_NO;i++)
  {
    if (dow_parport_table[i].dev != NULL)
    {
      parport_unregister_device(dow_parport_table[i].dev);
#ifdef DOW_DEBUG
      printk("Freeing %d\n",i);
#endif
    }
  }
#endif
  
}

